from base_tokenizer import BaseTokenizer
import json


class SilabiVocabGenerator:
    """
    Tokenizes text on kiswahili syllables
    """

    silabis = [
        "mbwa",
        "mbwe",
        "mbwi",
        "ndwa",
        "ndwe",
        "ndwi",
        "ngwa",
        "ngwe",
        "ngwi",
        "njwa",
        "njwe",
        "njwi",
        "nywa",
        "nywe",
        "shwa",
        "shwe",
        "shwi",
        "chwa",
        "chwe",
        "chwi",
        "pwa",
        "pwe",
        "pwi",
        "pwo",
        "swa",
        "swe",
        "swi",
        "twa",
        "twe",
        "twi",
        "zwa",
        "zwe",
        "zwi",
        "cha",
        "che",
        "chi",
        "cho",
        "chu",
        "dha",
        "dhe",
        "dhi",
        "dho",
        "dhu",
        "gha",
        "ghe",
        "ghi",
        "gho",
        "ghu",
        "kha",
        "khe",
        "kho",
        "khu",
        "mba",
        "mbe",
        "mbi",
        "mbo",
        "mbu",
        "nda",
        "nde",
        "ndi",
        "ndo",
        "ndu",
        "nga",
        "nge",
        "ngi",
        "ngo",
        "ngu",
        "ng’a",
        "ng’e",
        "ng’o",
        "nja",
        "nje",
        "nji",
        "njo",
        "nju",
        "nya",
        "nye",
        "nyi",
        "nyo",
        "nyu",
        "sha",
        "she",
        "shi",
        "sho",
        "shu",
        "tha",
        "the",
        "thi",
        "tho",
        "thu",
        "vya",
        "vye",
        "vyo",
        "bwa",
        "bwe",
        "bwi",
        "gwa",
        "gwe",
        "gwi",
        "jwa",
        "jwe",
        "jwi",
        "kwa",
        "kwe",
        "kwi",
        "lwa",
        "lwe",
        "lwi",
        "mwa",
        "mwe",
        "mwi",
        "nza",
        "nze",
        "nzi",
        "nzo",
        "nzu",
        "ba",
        "be",
        "bi",
        "bo",
        "bu",
        "da",
        "de",
        "di",
        "do",
        "du",
        "fa",
        "fe",
        "fi",
        "fo",
        "fu",
        "ga",
        "ge",
        "gi",
        "go",
        "gu",
        "ha",
        "he",
        "hi",
        "ho",
        "hu",
        "ja",
        "je",
        "ji",
        "jo",
        "ju",
        "ka",
        "ke",
        "ki",
        "ko",
        "ku",
        "la",
        "le",
        "li",
        "lo",
        "lu",
        "ma",
        "me",
        "mi",
        "mo",
        "mu",
        "na",
        "ne",
        "ni",
        "no",
        "nu",
        "pa",
        "pe",
        "pi",
        "po",
        "pu",
        "ra",
        "re",
        "ri",
        "ro",
        "ru",
        "sa",
        "se",
        "si",
        "so",
        "su",
        "ta",
        "te",
        "ti",
        "to",
        "va",
        "ve",
        "vi",
        "vo",
        "vu",
        "wa",
        "we",
        "wi",
        "wo",
        "wu",
        "ya",
        "ye",
        "yi",
        "yo",
        "yu",
        "vu",
        "za",
        "ze",
        "zi",
        "zo",
        "zu",
        "a",
        "e",
        "i",
        "o",
        "u",
    ]

    def __init__(self, vocab_prefix="swahili") -> None:
        self.vocab_prefix = vocab_prefix
        self.lookup = {}

    def generate_silabi_lookup(self):
        """
        Populates the lookup table with unknown,bos,eos,pad tokens
        Adds the utf-8 bytes to the lookup table
        Adds the silabis with capital letter, whitespace permutations
        """
        # add special tokens
        self.lookup["<unk>"] = 0
        self.lookup["<s>"] = 1
        self.lookup["<s/>"] = 2
        self.lookup["<pad>"] = 3

        # add utf-8 byte fallback tokens
        special_tokens_length = len(self.lookup)
        byte_values = list(range(256))
        for index, value in enumerate(byte_values):
            self.lookup[f"<0x{value:02x}>"] = index + special_tokens_length

        # add silabis and their permutation to the lookup
        counter = len(self.lookup)
        index = 0
        while index < len(self.silabis):
            silabi = self.silabis[index]
            capital_silabi = silabi.capitalize()
            whitespace_silabi = f"_{silabi}"
            whitespace_silabi_capital = f"_{capital_silabi}"

            permutations = [
                silabi,
                capital_silabi,
                whitespace_silabi,
                whitespace_silabi_capital,
            ]
            for permutation in permutations:
                self.lookup[permutation] = counter + 1
                counter += 1

            index += 1

        return self.lookup

    def write_to_file(self, file_path="swahili_vocab.json"):
        with open(file_path, "w") as f:
            json.dump(self.lookup, f, indent=4, ensure_ascii=False)





from abc import ABC, abstractmethod
from typing import List

class BaseTokenizer(ABC):


    def __call__(self, text: str):
        return self.tokenize(text)

    def tokenize(self, text: str) -> List[str]:
        raise NotImplementedError

    def tokens_to_ids(self, tokens: List[str]) -> List[int]:
        raise NotImplementedError

    def ids_to_tokens(self, ids : List[int]) -> List[str]:
        raise NotImplementedError

    def vocab_size(self) -> int:
        raise NotImplementedError



from hf_tokenizer import SilabiTokenizer

# Initialize the tokenizer
tokenizer = SilabiTokenizer()
# Encode a sample text
encoded_input = tokenizer("Hii ni mfano wa maandishi.")

# Decode the token ids back to text
decoded_text = tokenizer.decode(encoded_input['input_ids'])

print("Encoded Input:", encoded_input)
print("Decoded Text:", decoded_text)


from .base_tokenizer import BaseTokenizer
from .silabi_generator import SilabiVocabGenerator
import os
import re
import logging
import json
from typing import List

logger = logging.getLogger(__name__)


class KiswahiliTokenizer(BaseTokenizer):
    """
    Preferred
    Kiswahili Syllabic Tokenizer. With a byte fallback to utf-8.
    Generates a vocabulary file if not found.
    """

    def __init__(self) -> None:
        if not os.path.exists("silabi_vocab.json"):
            logger.debug("silabi_vocab.json not found, generating one")
            vocab = SilabiVocabGenerator()
            lookup = vocab.generate_silabi_lookup()
            vocab.write_to_file("silabi_vocab.json")
            self.vocab = lookup
        else:
            with open("silabi_vocab.json", "r") as f:
                self.vocab = json.load(f)
        self.reverse_vocab = {v: k for k, v in self.vocab.items()} # used for reverse lookup. from id to token
    
    def __call__(self, text: str):
        
        return self.tokenize(text)

    def tokenize(self, text: str) -> List[str]:
        """
        Tokenizes the text on the syllables
        """
        # add an initial whitespace to the text
        text = f" {text}"
        # replace the whitespace with _
        text = text.replace(" ", "_")
        syllables = []
        while len(text) > 0:
            found = False
            for i in range(1, 6):
                if text[:i] in self.vocab:
                    syllables.append(text[:i])
                    text = text[i:]
                    found = True
                    break
            if not found:
                syllables.append(text[0])
                text = text[1:]
        return syllables

    def tokens_to_ids(self, tokens: list) -> List[int]:
        """
        Converts tokens to their ids
        """
        ids = []
        for token in tokens:
            try:
                if token in self.vocab:
                    ids.append(self.vocab[token])
                else:
                    # utf byte fallback
                    ids.append(self.vocab[f"<0x{ord(token):02x}>"])
            except KeyError as _:
                ids.append(self.vocab["<unk>"])

        return ids

    def convert_byte_token(self, token: str) -> str:
        """
        Converts a byte token to its utf-8 representation
        """
        pattern = r"<0x([0-9A-Fa-f]{1,2})>"
        match = re.match(pattern, token)
        if match:
            return chr(int(match.group(1), 16))
        return token

    def ids_to_tokens(self, ids: List[int]) -> List[str]:
        """
        Converts ids to their tokens
        """

        return [self.convert_byte_token(self.reverse_vocab[id]) for id in ids]

    def vocab_size(self) -> int:
        """
        Returns the size of the vocabulary
        """
        return len(self.vocab)

    def tokens_to_sentence(self, tokens: List[str]) -> str:
        """
        Converts tokens to a sentence.
        Removes special tokens from the result sentence.
        """
        return "".join(tokens).replace("_", " ").replace('<unk>','').strip()

    def __repr__(self) -> str:
        return f"SilabiTokenizer()"

    @property
    def pad_token(self):
        return self.vocab["<pad>"]

    @property
    def bos_token(self):
        return self.vocab["<s>"]

    @property
    def eos_token(self):
        return self.vocab["<s/>"]

    @property
    def unk_token(self):
        return self.vocab["<unk>"]



if __name__ == "__main__":
    # Initialize the tokenizer
    tokenizer = KiswahiliTokenizer()
    # Encode a sample text
    encoded_input = tokenizer("Mfano Kenya.")
    print(encoded_input)


from typing import List
from transformers import PreTrainedTokenizer
from .kiswahili_tokenizer import KiswahiliTokenizer
import os
import json


class SilabiTokenizer(PreTrainedTokenizer):
    def __init__(self, *args, **kwargs):
        self.tokenizer = KiswahiliTokenizer()
        self.vocab = self.tokenizer.vocab
        self.ids_to_tokens = self.tokenizer.reverse_vocab
        super().__init__(*args, **kwargs)

    def _tokenize(self, text):
        return self.tokenizer.tokenize(text)

    def _convert_token_to_id(self, token):
        return self.tokenizer.tokens_to_ids([token])[0]

    def _convert_id_to_token(self, index):
        return self.tokenizer.convert_byte_token(self.ids_to_tokens.get(index, "<unk>"))

    def get_vocab(self):
        return self.vocab

    def vocab_size(self):
        return self.tokenizer.vocab_size()

    def save_vocabulary(self, save_directory=""):
        path = os.path.join(save_directory, "vocab.json")
        with open(path, "w", encoding="utf-8") as f:
            json.dump(self.vocab, f, ensure_ascii=False, indent=4)
        return (path,)

    def build_inputs_with_special_tokens(
        self, token_ids_0: List[int], token_ids_1: List[int] | None = None
    ) -> List[int]:

        return [self.vocab["<s>"]] + token_ids_0 + [self.vocab["<s/>"]]

    def get_special_tokens_mask(
        self, token_ids_0, token_ids_1=None, already_has_special_tokens=False
    ):
        return [1] + ([0] * len(token_ids_0)) + [1]

    def create_token_type_ids_from_sequences(self, token_ids_0, token_ids_1=None):
        return len(token_ids_0) * [0] + ([1] * len(token_ids_1) if token_ids_1 else [])



from setuptools import setup, find_packages

setup(
    name='silabi-tokenizer',
    version='0.3.0',
    description='A custom tokenizer for Swahili text using syllabic vocabulary with byte fallback.',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    author='Edwin Nguthiru',
    author_email='nguthiruedwin@gmail.com',
    url='https://github.com/nguthiru/silabi-tokenizer',
    license='MIT',
    packages=find_packages(),
    install_requires=[
        'transformers>=4.0.0',
        # Add any other dependencies here
    ],
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6',
)
